@@ -56,7 +56,7 @@ struct steve_poll_waiter {
 
 struct steve_process {
 	int pid_fd{-1};
-	size_t tokens_held{0};
+	ssize_t tokens_held{0};
 	std::unique_ptr<struct event, std::function<void(struct event*)>> event;
 
 	~steve_process() {
@@ -156,6 +156,31 @@ static void steve_open(fuse_req_t req, struct fuse_file_info *fi)
 	static_assert(sizeof(fi->fh) >= sizeof(context->pid));
 	fi->fh = context->pid;
 
+	if (state->verbose) {
+		char cmdline[128] = {};
+
+		std::string path = std::format("/proc/{}/cmdline", fi->fh);
+		FILE *cmdline_file = fopen(path.c_str(), "r");
+		if (cmdline_file) {
+			size_t rd = fread(cmdline, 1, sizeof(cmdline) - 1, cmdline_file);
+			if (rd > 0) {
+				/* replace all NULs with spaces, except for the final one */
+				for (size_t i = 0; i < rd - 1; ++i) {
+					if (cmdline[i] == 0)
+						cmdline[i] = ' ';
+				}
+				/* ensure a NUL, in case it was truncated */
+				cmdline[rd] = 0;
+			}
+			fclose(cmdline_file);
+		}
+
+		if (cmdline[0]) {
+			std::print(stderr, "Device open by PID {} ({})\n", fi->fh, cmdline);
+		} else
+			std::print(stderr, "Device open by PID {} (process name unknown)\n", fi->fh);
+	}
+
 	int pid_fd;
 	if (state->processes.find(fi->fh) != state->processes.end()) {
 		assert(state->processes[fi->fh].pid_fd != -1);
@@ -188,9 +213,6 @@ static void steve_open(fuse_req_t req, struct fuse_file_info *fi)
 		state->processes[fi->fh].event = std::move(pidfd_event);
 	}
 
-	if (state->verbose)
-		std::print(stderr, "Device open by PID {} (pidfd {})\n", fi->fh, pid_fd);
-
 	fuse_reply_open(req, fi);
 }
 
@@ -261,11 +283,24 @@ static void steve_write(
 		fuse_reply_err(req, EIO);
 		return;
 	}
+	if (size > SSIZE_MAX) {
+		std::print(stderr, "Warning: process {} tried to return more than SSIZE_MAX tokens\n",
+				fi->fh);
+		fuse_reply_err(req, EFBIG);
+		return;
+	}
 
-	if (state->processes[fi->fh].tokens_held < size) {
+	/* workaround for https://github.com/medek/nasm-rs/issues/44 */
+	if (state->processes[fi->fh].tokens_held == 0 && size == 1) {
+		std::print(stderr, "Warning: process {} pre-released an unacquired token, please report a bug upstream\n",
+				fi->fh);
+	} else if (state->processes[fi->fh].tokens_held < static_cast<ssize_t>(size)) {
 		std::print(stderr, "Warning: process {} tried to return {} tokens while holding only {} tokens, capping\n",
 				fi->fh, size, state->processes[fi->fh].tokens_held);
-		size = state->processes[fi->fh].tokens_held;
+		if (state->processes[fi->fh].tokens_held < 0)
+			size = 0;
+		else
+			size = state->processes[fi->fh].tokens_held;
 	}
 	if (size == 0) {
 		fuse_reply_err(req, ENOSPC);
@@ -454,7 +489,7 @@ int main(int argc, char **argv)
 		return 1;
 	}
 
-	std::string mountpoint = "/dev/fd/" + std::to_string(cuse_fd);
+	std::string mountpoint = std::format("/dev/fd/{}", cuse_fd);
 	if (fuse_session_mount(session.get(), mountpoint.c_str()) == -1) {
 		std::print(stderr, "failed to mount the filesystem");
 		return 1;
